var Base = require('basejs')
var ActionCable = require('../action_cable.js')
var INTERNAL_JSON = require('../internal.js')
var ConnectionMonitor = require('./connection_monitor.js')

var message_types = INTERNAL_JSON.message_types
var supportedProtocols = INTERNAL_JSON.protocols

var bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; },
  indexOf = [].indexOf || function(item) { for (var i = 0, l = this.length; i < l; i++) { if (i in this && this[i] === item) return i; } return -1; },
  slice = [].slice;

var Connection = Base.extend({
  constructor: function (consumer) {
    this.consumer = consumer;
    this.open = bind(this.open, this);
    this.subscriptions = this.consumer.subscriptions;
    this.monitor = new ConnectionMonitor(this).self;
    this.disconnected = true;
    this.self = this
  },

  reopenDelay: 500,

  send: function(data) {
    if (this.isOpen()) {
      this.webSocket.send(JSON.stringify(data));
      return true;
    } else {
      return false;
    }
  },

  open: function () {
    if (this.isActive()) {
      ActionCable.log("Attempted to open WebSocket, but existing socket is " + (this.getState()));
      throw new Error("Existing connection must be closed before opening");
    } else {
      ActionCable.log("Opening WebSocket, current state is " + (this.getState()) + ", subprotocols: " + protocols);
      if (this.webSocket != null) {
        this.uninstallEventHandlers();
      }
      this.webSocket = new WebSocket(this.consumer.url, protocols);
      this.installEventHandlers();
      this.monitor.start();
      return true;
    }
  },

  close: function (arg) {
    var allowReconnect, ref;
    allowReconnect = (arg != null ? arg : {
      allowReconnect: true
    }).allowReconnect;
    if (!allowReconnect) {
      this.monitor.stop();
    }
    if (this.isActive()) {
      return (ref = this.webSocket) != null ? ref.close() : void 0;
    }
  },

  reopen: function () {
    var error;
    ActionCable.log("Reopening WebSocket, current state is " + (this.getState()));
    if (this.isActive()) {
      try {
        return this.close();
      } catch (_error) {
        error = _error;
        return ActionCable.log("Failed to reopen WebSocket", error);
      } finally {
        ActionCable.log("Reopening WebSocket in " + this.constructor.reopenDelay + "ms");
        setTimeout(this.open, this.constructor.reopenDelay);
      }
    } else {
      return this.open();
    }
  },

  getProtocol: function() {
    var ref;
    return (ref = this.webSocket) != null ? ref.protocol : void 0;
  },

  isOpen: function() {
    return this.isState("open");
  },

  isActive: function() {
    return this.isState("open", "connecting");
  },

  isProtocolSupported: function () {
    var ref;
    return ref = this.getProtocol(), indexOf.call(supportedProtocols, ref) >= 0;
  },

  isState: function() {
    var ref, states;
    states = 1 <= arguments.length ? slice.call(arguments, 0) : [];
    return ref = this.getState(), indexOf.call(states, ref) >= 0;
  },

  getState: function() {
    var ref, state, value;
    for (state in WebSocket) {
      value = WebSocket[state];
      if (value === ((ref = this.webSocket) != null ? ref.readyState : void 0)) {
        return state.toLowerCase();
      }
    }
    return null;
  },

  installEventHandlers: function() {
    var eventName, handler;
    for (eventName in this.events) {
      handler = this.events[eventName].bind(this);
      this.webSocket["on" + eventName] = handler;
    }
  },

  uninstallEventHandlers: function () {
    var eventName;
    for (eventName in this.events) {
      this.webSocket["on" + eventName] = function() {};
    }
  },

  events: function () {
    return {
      message: function(event) {
        var identifier, message, ref, type;
        if (!this.isProtocolSupported()) {
          return;
        }
        ref = JSON.parse(event.data), identifier = ref.identifier, message = ref.message, type = ref.type;
        switch (type) {
          case message_types.welcome:
            this.monitor.recordConnect();
            return this.subscriptions.reload();
          case message_types.ping:
            return this.monitor.recordPing();
          case message_types.confirmation:
            return this.subscriptions.notify(identifier, "connected");
          case message_types.rejection:
            return this.subscriptions.reject(identifier);
          default:
            return this.subscriptions.notify(identifier, "received", message);
        }
      },
      open: function() {
        ActionCable.log("WebSocket onopen event, using '" + (this.getProtocol()) + "' subprotocol");
        this.disconnected = false;
        if (!this.isProtocolSupported()) {
          ActionCable.log("Protocol is unsupported. Stopping monitor and disconnecting.");
          return this.close({
            allowReconnect: false
          });
        }
      },
      close: function(event) {
        ActionCable.log("WebSocket onclose event");
        if (this.disconnected) {
          return;
        }
        this.disconnected = true;
        this.monitor.recordDisconnect();
        return this.subscriptions.notifyAll("disconnected", {
          willAttemptReconnect: this.monitor.isRunning()
        });
      },
      error: function() {
        return ActionCable.log("WebSocket onerror event");
      }
    }
  }

});

module.exports = Connection